@@ -35,29 +35,48 @@ namespace Gwenview
 
 struct RecursiveDirModelPrivate {
     KDirLister* mDirLister;
-    KFileItemList mList;
 
-    int indexForUrl(const KUrl &url) const
+    int rowForUrl(const KUrl &url) const
     {
-        int row = 0;
-        KFileItemList::ConstIterator it = mList.begin(), end = mList.end();
-        for (; it != end; ++it , ++row) {
-            if (it->url() == url) {
-                return row;
-            }
-        }
-        return -1;
+        return mRowForUrl.value(url, -1);
     }
 
     void removeAt(int row)
     {
-        mList.removeAt(row);
+        KFileItem item = mList.takeAt(row);
+        mRowForUrl.remove(item.url());
+
+        // Decrease row value for all urls after the one we removed
+        // ("row" now points to the item after the one we removed since we used takeAt)
+        const int count = mList.count();
+        for (; row < count; ++row) {
+            KUrl url = mList.at(row).url();
+            mRowForUrl[url]--;
+        }
     }
 
     void addItem(const KFileItem& item)
     {
+        mRowForUrl.insert(item.url(), mList.count());
         mList.append(item);
     }
+
+    void clear()
+    {
+        mRowForUrl.clear();
+        mList.clear();
+    }
+
+    // RecursiveDirModel can only access mList through this read-only getter.
+    // This ensures it cannot introduce inconsistencies between mList and mRowForUrl.
+    const KFileItemList& list() const
+    {
+        return mList;
+    }
+
+private:
+    KFileItemList mList;
+    QHash<KUrl, int> mRowForUrl;
 };
 
 RecursiveDirModel::RecursiveDirModel(QObject* parent)
@@ -90,7 +109,7 @@ KUrl RecursiveDirModel::url() const
 void RecursiveDirModel::setUrl(const KUrl& url)
 {
     beginResetModel();
-    d->mList.clear();
+    d->clear();
     endResetModel();
     d->mDirLister->openUrl(url);
 }
@@ -100,7 +119,7 @@ int RecursiveDirModel::rowCount(const QModelIndex& parent) const
     if (parent.isValid()) {
         return 0;
     } else {
-        return d->mList.count();
+        return d->list().count();
     }
 }
 
@@ -109,7 +128,7 @@ QVariant RecursiveDirModel::data(const QModelIndex& index, int role) const
     if (index.parent().isValid()) {
         return QVariant();
     }
-    KFileItem item = d->mList.value(index.row());
+    KFileItem item = d->list().value(index.row());
     if (item.isNull()) {
         kWarning() << "Invalid row" << index.row();
         return QVariant();
@@ -131,17 +150,25 @@ QVariant RecursiveDirModel::data(const QModelIndex& index, int role) const
 void RecursiveDirModel::slotItemsAdded(const KUrl&, const KFileItemList& newList)
 {
     QList<KUrl> dirUrls;
+    KFileItemList fileList;
     Q_FOREACH(const KFileItem& item, newList) {
         if (item.isFile()) {
-            if (d->indexForUrl(item.url()) == -1) {
-                beginInsertRows(QModelIndex(), d->mList.count(), d->mList.count());
-                d->addItem(item);
-                endInsertRows();
+            if (d->rowForUrl(item.url()) == -1) {
+                fileList << item;
             }
         } else {
             dirUrls << item.url();
         }
     }
+
+    if (!fileList.isEmpty()) {
+        beginInsertRows(QModelIndex(), d->list().count(), d->list().count() + fileList.count());
+        Q_FOREACH(const KFileItem& item, fileList) {
+            d->addItem(item);
+        }
+        endInsertRows();
+    }
+
     Q_FOREACH(const KUrl& url, dirUrls) {
         d->mDirLister->openUrl(url, KDirLister::Keep);
     }
@@ -153,7 +180,7 @@ void RecursiveDirModel::slotItemsDeleted(const KFileItemList& list)
         if (item.isDir()) {
             continue;
         }
-        int row = d->indexForUrl(item.url());
+        int row = d->rowForUrl(item.url());
         if (row == -1) {
             kWarning() << "Received itemsDeleted for an unknown item: this should not happen!";
             continue;
@@ -166,19 +193,19 @@ void RecursiveDirModel::slotItemsDeleted(const KFileItemList& list)
 
 void RecursiveDirModel::slotCleared()
 {
-    if (d->mList.isEmpty()) {
+    if (d->list().isEmpty()) {
         return;
     }
-    beginRemoveRows(QModelIndex(), 0, d->mList.count() - 1);
-    d->mList.clear();
-    endRemoveRows();
+    beginResetModel();
+    d->clear();
+    endResetModel();
 }
 
 void RecursiveDirModel::slotDirCleared(const KUrl& dirUrl)
 {
     int row;
-    for (row = d->mList.count() - 1; row >= 0; --row) {
-        const KUrl url = d->mList.at(row).url();
+    for (row = d->list().count() - 1; row >= 0; --row) {
+        const KUrl url = d->list().at(row).url();
         if (dirUrl.isParentOf(url)) {
             beginRemoveRows(QModelIndex(), row, row);
             d->removeAt(row);
